@@ -1,15 +1,25 @@
+"""
+$description Live TV channels and video on-demand service from MRT, a North Macedonian public, state-owned broadcaster.
+$url play.mrt.com.mk
+$type live, vod
+$region North Macedonia
+"""
+
 import logging
 import re
 
-from streamlink.plugin import Plugin
+from streamlink.plugin import Plugin, pluginmatcher
 from streamlink.plugin.api import validate
-from streamlink.stream import HLSStream
+from streamlink.stream.hls import HLSStream
+
 
 log = logging.getLogger(__name__)
 
 
+@pluginmatcher(
+    re.compile(r"https?://play\.mrt\.com\.mk/(live|play)/"),
+)
 class MRTmk(Plugin):
-    url_re = re.compile(r"""https?://play\.mrt\.com\.mk/(live|play)/""")
     file_re = re.compile(r"""(?P<url>https?://vod-[\d\w]+\.interspace\.com[^"',]+\.m3u8[^"',]*)""")
 
     stream_schema = validate.Schema(
@@ -23,10 +33,6 @@ class MRTmk(Plugin):
         ),
     )
 
-    @classmethod
-    def can_handle_url(cls, url):
-        return cls.url_re.match(url) is not None
-
     def _get_streams(self):
         res = self.session.http.get(self.url)
         stream_urls = self.stream_schema.validate(res.text)
