@@ -1,8 +1,11 @@
+from __future__ import annotations
+
 import base64
 import re
+from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine, Mapping
 from contextlib import asynccontextmanager
 from dataclasses import dataclass
-from typing import Any, AsyncGenerator, Awaitable, Callable, Coroutine, List, Mapping, Optional, Set
+from typing import TYPE_CHECKING, Any
 
 import trio
 
@@ -13,13 +16,14 @@ from streamlink.webbrowser.cdp.exceptions import CDPError
 from streamlink.webbrowser.chromium import ChromiumWebbrowser
 
 
-try:
-    from typing import Self, TypeAlias  # type: ignore[attr-defined]
-except ImportError:  # pragma: no cover
-    from typing_extensions import Self, TypeAlias
+if TYPE_CHECKING:
+    try:
+        from typing import Self, TypeAlias  # type: ignore[attr-defined]
+    except ImportError:
+        from typing_extensions import Self, TypeAlias
 
 
-TRequestHandlerCallable: TypeAlias = Callable[["CDPClientSession", fetch.RequestPaused], Awaitable]
+TRequestHandlerCallable: TypeAlias = "Callable[[CDPClientSession, fetch.RequestPaused], Awaitable]"
 
 
 _re_url_pattern_wildcard = re.compile(r"(.+?)?(\\+)?([*?])")
@@ -65,7 +69,7 @@ class RequestPausedHandler:
 class CMRequestProxy:
     body: str
     response_code: int
-    response_headers: Optional[Mapping[str, str]]
+    response_headers: Mapping[str, str] | None
 
 
 class CDPClient:
@@ -86,21 +90,23 @@ class CDPClient:
     ``streamlink.webbrowser.cdp.devtools`` package, but be aware that only a subset of the available domains is supported.
     """
 
-    def __init__(self, cdp_connection: CDPConnection, nursery: trio.Nursery):
+    def __init__(self, session: Streamlink, cdp_connection: CDPConnection, nursery: trio.Nursery, headless: bool):
+        self.streamlink = session
         self.cdp_connection = cdp_connection
         self.nursery = nursery
+        self.headless = headless
 
     @classmethod
     def launch(
         cls,
         session: Streamlink,
         runner: Callable[[Self], Coroutine],
-        executable: Optional[str] = None,
-        timeout: Optional[float] = None,
-        cdp_host: Optional[str] = None,
-        cdp_port: Optional[int] = None,
-        cdp_timeout: Optional[float] = None,
-        headless: Optional[bool] = None,
+        executable: str | None = None,
+        timeout: float | None = None,
+        cdp_host: str | None = None,
+        cdp_port: int | None = None,
+        cdp_timeout: float | None = None,
+        headless: bool | None = None,
     ) -> Any:
         """
         Start a new :mod:`trio` runloop and do the following things:
@@ -166,37 +172,42 @@ class CDPClient:
             ) as cdp_client:
                 return await runner(cdp_client)
 
-        return trio.run(run_wrapper)
+        return trio.run(run_wrapper, strict_exception_groups=True)
 
     @classmethod
     @asynccontextmanager
     async def run(
         cls,
         session: Streamlink,
-        executable: Optional[str] = None,
-        timeout: Optional[float] = None,
-        cdp_host: Optional[str] = None,
-        cdp_port: Optional[int] = None,
-        cdp_timeout: Optional[float] = None,
-        headless: bool = True,
+        executable: str | None = None,
+        timeout: float | None = None,
+        cdp_host: str | None = None,
+        cdp_port: int | None = None,
+        cdp_timeout: float | None = None,
+        headless: bool = False,
     ) -> AsyncGenerator[Self, None]:
-        webbrowser = ChromiumWebbrowser(executable=executable, host=cdp_host, port=cdp_port, headless=headless)
+        webbrowser = ChromiumWebbrowser(executable=executable, host=cdp_host, port=cdp_port)
         nursery: trio.Nursery
-        async with webbrowser.launch(timeout=timeout) as nursery:
+        async with webbrowser.launch(headless=headless, timeout=timeout) as nursery:
             websocket_url = webbrowser.get_websocket_url(session)
             cdp_connection: CDPConnection
             async with CDPConnection.create(websocket_url, timeout=cdp_timeout) as cdp_connection:
-                yield cls(cdp_connection, nursery)
+                yield cls(session, cdp_connection, nursery, headless)
 
     @asynccontextmanager
-    async def session(self, fail_unhandled_requests: bool = False) -> AsyncGenerator["CDPClientSession", None]:
+    async def session(
+        self,
+        fail_unhandled_requests: bool = False,
+        max_buffer_size: int | None = None,
+    ) -> AsyncGenerator[CDPClientSession, None]:
         """
         Create a new CDP session on an empty target (browser tab).
 
         :param fail_unhandled_requests: Whether network requests which are not matched by any request handlers should fail.
+        :param max_buffer_size: Optional size of the send/receive memory channel for paused HTTP requests/responses.
         """
         cdp_session = await self.cdp_connection.new_target()
-        yield CDPClientSession(self, cdp_session, fail_unhandled_requests)
+        yield CDPClientSession(self, cdp_session, fail_unhandled_requests, max_buffer_size)
 
 
 class CDPClientSession:
@@ -212,12 +223,14 @@ class CDPClientSession:
         cdp_client: CDPClient,
         cdp_session: CDPSession,
         fail_unhandled_requests: bool = False,
+        max_buffer_size: int | None = None,
     ):
         self.cdp_client = cdp_client
         self.cdp_session = cdp_session
         self._fail_unhandled = fail_unhandled_requests
-        self._request_handlers: List[RequestPausedHandler] = []
-        self._requests_handled: Set[str] = set()
+        self._request_handlers: list[RequestPausedHandler] = []
+        self._requests_handled: set[str] = set()
+        self._max_buffer_size = max_buffer_size
 
     def add_request_handler(
         self,
@@ -242,7 +255,7 @@ class CDPClientSession:
         )
 
     @asynccontextmanager
-    async def navigate(self, url: str, referrer: Optional[str] = None) -> AsyncGenerator[page.FrameId, None]:
+    async def navigate(self, url: str, referrer: str | None = None) -> AsyncGenerator[page.FrameId, None]:
         """
         Async context manager for opening the URL with an optional referrer and starting the optional interception
         of network requests and responses.
@@ -268,6 +281,9 @@ class CDPClientSession:
         async with trio.open_nursery() as nursery:
             nursery.start_soon(self._on_target_detached_from_target)
 
+            if self.cdp_client.headless:
+                await self._update_user_agent()
+
             if request_patterns:
                 nursery.start_soon(self._on_fetch_request_paused)
                 await self.cdp_session.send(fetch.enable(request_patterns, True))
@@ -275,7 +291,9 @@ class CDPClientSession:
             await self.cdp_session.send(page.enable())
 
             try:
-                frame_id, _loader_id, error = await self.cdp_session.send(page.navigate(url=url, referrer=referrer))
+                frame_id, _loader_id, error, _is_download = await self.cdp_session.send(
+                    page.navigate(url=url, referrer=referrer),
+                )
                 if error:
                     raise CDPError(f"Navigation error: {error}")
 
@@ -295,7 +313,7 @@ class CDPClientSession:
             if frame_stopped_loading.frame_id == frame_id:
                 return
 
-    async def evaluate(self, expression: str, await_promise: bool = True, timeout: Optional[float] = None) -> Any:
+    async def evaluate(self, expression: str, await_promise: bool = True, timeout: float | None = None) -> Any:
         """
         Evaluate an optionally async JavaScript expression and return its result.
 
@@ -321,53 +339,59 @@ class CDPClientSession:
     async def continue_request(
         self,
         request: fetch.RequestPaused,
-        url: Optional[str] = None,
-        method: Optional[str] = None,
-        post_data: Optional[str] = None,
-        headers: Optional[Mapping[str, str]] = None,
+        url: str | None = None,
+        method: str | None = None,
+        post_data: str | None = None,
+        headers: Mapping[str, str] | None = None,
     ):
         """
         Continue a request and optionally override the request method, URL, POST data or request headers.
         """
-        await self.cdp_session.send(fetch.continue_request(
-            request_id=request.request_id,
-            url=url,
-            method=method,
-            post_data=base64.b64encode(post_data.encode()).decode() if post_data is not None else None,
-            headers=self._headers_entries_from_mapping(headers),
-        ))
+        await self.cdp_session.send(
+            fetch.continue_request(
+                request_id=request.request_id,
+                url=url,
+                method=method,
+                post_data=base64.b64encode(post_data.encode()).decode() if post_data is not None else None,
+                headers=self._headers_entries_from_mapping(headers),
+            ),
+        )
         self._requests_handled.add(request.request_id)
 
     async def fail_request(
         self,
         request: fetch.RequestPaused,
-        error_reason: Optional[str] = None,
+        error_reason: str | None = None,
     ):
         """
         Let a request fail, with an optional error reason which defaults to ``BlockedByClient``.
         """
-        await self.cdp_session.send(fetch.fail_request(
-            request_id=request.request_id,
-            error_reason=network.ErrorReason(error_reason or network.ErrorReason.BLOCKED_BY_CLIENT),
-        ))
+        await self.cdp_session.send(
+            fetch.fail_request(
+                request_id=request.request_id,
+                error_reason=network.ErrorReason(error_reason or network.ErrorReason.BLOCKED_BY_CLIENT),
+            ),
+        )
         self._requests_handled.add(request.request_id)
 
     async def fulfill_request(
         self,
         request: fetch.RequestPaused,
         response_code: int = 200,
-        response_headers: Optional[Mapping[str, str]] = None,
-        body: Optional[str] = None,
+        response_headers: Mapping[str, str] | None = None,
+        body: str | None = None,
     ) -> None:
         """
         Fulfill a response and override its status code, headers and body.
         """
-        await self.cdp_session.send(fetch.fulfill_request(
-            request_id=request.request_id,
-            response_code=response_code,
-            response_headers=self._headers_entries_from_mapping(response_headers),
-            body=base64.b64encode(body.encode()).decode() if body is not None else None,
-        ))
+        await self.cdp_session.send(
+            fetch.fulfill_request(
+                request_id=request.request_id,
+                response_code=response_code,
+                response_headers=self._headers_entries_from_mapping(response_headers),
+                body=base64.b64encode(body.encode()).decode() if body is not None else None,
+            ),
+        )
         self._requests_handled.add(request.request_id)
 
     @asynccontextmanager
@@ -375,7 +399,7 @@ class CDPClientSession:
         self,
         request: fetch.RequestPaused,
         response_code: int = 200,
-        response_headers: Optional[Mapping[str, str]] = None,
+        response_headers: Mapping[str, str] | None = None,
     ) -> AsyncGenerator[CMRequestProxy, None]:
         """
         Async context manager wrapper around :meth:`fulfill_request()` which retrieves the response body,
@@ -398,19 +422,18 @@ class CDPClientSession:
         )
 
     @staticmethod
-    def _headers_entries_from_mapping(headers: Optional[Mapping[str, str]]):
-        return None if headers is None else [
-            fetch.HeaderEntry(name=name, value=value)
-            for name, value in headers.items()
-        ]
+    def _headers_entries_from_mapping(headers: Mapping[str, str] | None):
+        return None if headers is None else [fetch.HeaderEntry(name=name, value=value) for name, value in headers.items()]
 
     async def _on_target_detached_from_target(self) -> None:
+        detached_from_target: target.DetachedFromTarget
         async for detached_from_target in self.cdp_client.cdp_connection.listen(target.DetachedFromTarget):
             if detached_from_target.session_id == self.cdp_session.session_id:
                 raise CDPError("Target has been detached")
 
     async def _on_fetch_request_paused(self) -> None:
-        async for request in self.cdp_session.listen(fetch.RequestPaused):
+        request: fetch.RequestPaused
+        async for request in self.cdp_session.listen(fetch.RequestPaused, max_buffer_size=self._max_buffer_size):
             for handler in self._request_handlers:
                 if not handler.matches(request):
                     continue
@@ -422,3 +445,46 @@ class CDPClientSession:
                     await self.fail_request(request)
                 else:
                     await self.continue_request(request)
+
+    async def _update_user_agent(self) -> None:
+        user_agent: str = await self.evaluate("navigator.userAgent", await_promise=False)
+        if not user_agent:  # pragma: no cover
+            raise CDPError("Could not read navigator.userAgent value")
+        user_agent = re.sub(r"Headless", "", user_agent, flags=re.IGNORECASE)
+        await self.cdp_session.send(network.set_user_agent_override(user_agent=user_agent))
+
+    async def apply_cookies(self) -> None:
+        """
+        Copy all cookies from Streamlink's HTTP session to the CDP session.
+        """
+        cookies = [
+            network.CookieParam(
+                name=cookie.name,
+                value=cookie.value or "",
+                domain=cookie.domain,
+                path=cookie.path,
+                expires=network.TimeSinceEpoch(cookie.expires) if cookie.expires is not None else None,
+                secure=cookie.secure,
+            )
+            for cookie in self.cdp_client.streamlink.http.cookies
+        ]
+        await self.cdp_session.send(
+            network.set_cookies(cookies=cookies),
+        )
+
+    async def retrieve_cookies(self) -> None:
+        """
+        Copy all cookies from the CDP session to Streamlink's HTTP session.
+        """
+        cookies = await self.cdp_session.send(
+            network.get_cookies(),
+        )
+        for cookie in cookies:
+            self.cdp_client.streamlink.http.cookies.set(
+                name=cookie.name,
+                value=cookie.value,
+                domain=cookie.domain,
+                path=cookie.path,
+                expires=None if cookie.expires == -1 else cookie.expires,
+                secure=cookie.secure,
+            )
